import oopt_gnpy_libyang as ly
import os
from pathlib import Path
import pytest

@pytest.fixture(scope='function')
def context_no_libyang():
    return ly.Context(str(Path(os.environ.get('CMAKE_CURRENT_SOURCE_DIR', os.getcwd())) / 'tests' / 'yang'),
                      ly.ContextOptions.AllImplemented | ly.ContextOptions.NoYangLibrary)

@pytest.fixture(scope='function')
def context_with_modules(context_no_libyang):
    '''A context with all relevant YANG modules added and implemented'''
    c = context_no_libyang
    c.load_module('iana-if-type')
    m = c.load_module('ietf-interfaces')
    m.set_implemented_all_features()
    c.load_module('ietf-ip')
    c.load_module('ietf-hardware')
    return c



import pytest
import oopt_gnpy_libyang as ly
from pathlib import Path

def test_no_module(context_no_libyang):
    with pytest.raises(ly.Error, match="Can't load module 'yay'"):
        context_no_libyang.load_module('yay')

def test_empty():
    c = ly.Context()
    with pytest.raises(ly.Error, match="Can't load module 'yay'"):
        c.load_module('yay')

def test_features(context_no_libyang):
    context_no_libyang.load_module('iana-if-type')
    m = context_no_libyang.load_module('ietf-interfaces')
    with pytest.raises(ly.Error, match="Couldn't set module 'ietf-interfaces' to implemented: LY_EINVAL"):
       m.set_implemented_with_features(['arbotrary-names'])
    errors = [(e.level, e.code, e.message, e.path, e.validation_code, e.app_tag) for e in context_no_libyang.errors()]
    assert errors == [
        (ly.LogLevel.Error, ly.ErrorCode.InvalidValue, 'Feature "arbotrary-names" not found in module "ietf-interfaces".',
         None, ly.ValidationErrorCode.Success, None)
    ]
    m.set_implemented_with_features(['arbitrary-names'])
    for feature in m.features:
        assert m.feature_enabled(feature.name) == (feature.name == 'arbitrary-names')
    m.set_implemented_all_features()
    for feature in m.features:
        assert m.feature_enabled(feature.name)

def test_no_shared_errors(context_no_libyang):
    assert context_no_libyang.errors() == []

def test_explicit_loading(context_no_libyang):
    TEST_ROOT = Path(__file__).parent
    with pytest.raises(ly.Error, match="Can't parse module: LY_ESYS"):
        context_no_libyang.parse_module(TEST_ROOT / 'no-such-path.yang', ly.SchemaFormat.YANG)
    m1 = context_no_libyang.parse_module(TEST_ROOT / 'yang' / 'iana-if-type@2017-01-19.yang', ly.SchemaFormat.YANG)
    assert m1.name == 'iana-if-type'
    m2 = context_no_libyang.parse_module((TEST_ROOT / 'yang' / 'iana-hardware@2018-03-13.yang').read_text(), ly.SchemaFormat.YANG)
    assert m2.name == 'iana-hardware'

def test_version():
    assert len(ly.libyang_version_info()) == 3
    assert ly.libyang_version_info()[0] == 2
    assert ly.libyang_version()[0:2] == '2.'



import pytest
import oopt_gnpy_libyang as ly

def test_ietf_interfaces(context_with_modules):
    blob = '''{
  "ietf-interfaces:interfaces": {
    "interface": [
      {
        "name": "lo",
        "type": "iana-if-type:softwareLoopback",
        "ietf-ip:ipv4": {
          "address": [
            {
              "ip": "127.0.0.1",
              "prefix-length": 8
            }
          ]
        },
        "ietf-ip:ipv6": {
          "address": [
            {
              "ip": "::1",
              "prefix-length": 128
            }
          ]
        }
      },
      {
        "name": "eth0",
        "type": "iana-if-type:ethernetCsmacd"
      },
      {
        "name": "eth1",
        "type": "iana-if-type:ethernetCsmacd"
      },
      {
        "name": "br0",
        "type": "iana-if-type:bridge",
        "ietf-ip:ipv4": {
          "address": [
            {
              "ip": "10.0.0.2",
              "prefix-length": 24
            }
          ],
          "neighbor": [
            {
              "ip": "10.0.0.1",
              "link-layer-address": "cc:2d:e0:52:30:9c"
            },
            {
              "ip": "10.0.1.1",
              "link-layer-address": "cc:2d:e0:52:30:9c"
            }
          ]
        },
        "ietf-ip:ipv6": {
          "address": [
            {
              "ip": "fe80::211:17ff:fe01:61",
              "prefix-length": 64
            }
          ]
        }
      }
    ]
  },
  "ietf-hardware:hardware": {
  }
}
    '''
    data = context_with_modules.parse_data(blob, ly.DataFormat.JSON, ly.ParseOptions.Strict | ly.ParseOptions.Ordered, ly.ValidationOptions.Present | ly.ValidationOptions.NoState)
    assert data.path == '/ietf-interfaces:interfaces'

    assert [x.path for x in data.siblings()] == ['/ietf-interfaces:interfaces', '/ietf-hardware:hardware']
    assert [inner.path for top in data.siblings() for inner in top.childrenDfs()] == [
        '/ietf-interfaces:interfaces',
        "/ietf-interfaces:interfaces/interface[name='lo']",
        "/ietf-interfaces:interfaces/interface[name='lo']/name",
        "/ietf-interfaces:interfaces/interface[name='lo']/type",
        "/ietf-interfaces:interfaces/interface[name='lo']/enabled",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv4",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv4/enabled",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv4/forwarding",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv4/address[ip='127.0.0.1']",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv4/address[ip='127.0.0.1']/ip",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv4/address[ip='127.0.0.1']/prefix-length",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv6",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv6/enabled",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv6/forwarding",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv6/address[ip='::1']",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv6/address[ip='::1']/ip",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv6/address[ip='::1']/prefix-length",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv6/dup-addr-detect-transmits",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv6/autoconf",
        "/ietf-interfaces:interfaces/interface[name='lo']/ietf-ip:ipv6/autoconf/create-global-addresses",
        "/ietf-interfaces:interfaces/interface[name='eth0']",
        "/ietf-interfaces:interfaces/interface[name='eth0']/name",
        "/ietf-interfaces:interfaces/interface[name='eth0']/type",
        "/ietf-interfaces:interfaces/interface[name='eth0']/enabled",
        "/ietf-interfaces:interfaces/interface[name='eth1']",
        "/ietf-interfaces:interfaces/interface[name='eth1']/name",
        "/ietf-interfaces:interfaces/interface[name='eth1']/type",
        "/ietf-interfaces:interfaces/interface[name='eth1']/enabled",
        "/ietf-interfaces:interfaces/interface[name='br0']",
        "/ietf-interfaces:interfaces/interface[name='br0']/name",
        "/ietf-interfaces:interfaces/interface[name='br0']/type",
        "/ietf-interfaces:interfaces/interface[name='br0']/enabled",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/enabled",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/forwarding",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/address[ip='10.0.0.2']",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/address[ip='10.0.0.2']/ip",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/address[ip='10.0.0.2']/prefix-length",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/neighbor[ip='10.0.0.1']",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/neighbor[ip='10.0.0.1']/ip",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/neighbor[ip='10.0.0.1']/link-layer-address",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/neighbor[ip='10.0.1.1']",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/neighbor[ip='10.0.1.1']/ip",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv4/neighbor[ip='10.0.1.1']/link-layer-address",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv6",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv6/enabled",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv6/forwarding",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv6/address[ip='fe80::211:17ff:fe01:61']",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv6/address[ip='fe80::211:17ff:fe01:61']/ip",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv6/address[ip='fe80::211:17ff:fe01:61']/prefix-length",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv6/dup-addr-detect-transmits",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv6/autoconf",
        "/ietf-interfaces:interfaces/interface[name='br0']/ietf-ip:ipv6/autoconf/create-global-addresses",
        '/ietf-hardware:hardware',
    ]
    assert [inner.path for top in data.siblings() for inner in top.immediateChildren()] == [
        f"/ietf-interfaces:interfaces/interface[name='{if_}']" for if_ in ('lo', 'eth0', 'eth1', 'br0')
    ]

    # quoting doesn't matter
    assert data["interface[name='lo']"].path == "/ietf-interfaces:interfaces/interface[name='lo']"
    assert data['interface[name="lo"]'].path == "/ietf-interfaces:interfaces/interface[name='lo']"

    with pytest.raises(KeyError, match='No such data node below.*'):
        data['interface[name="ZZZ"]']

    with pytest.raises(KeyError, match="Too many nodes below '/ietf-interfaces:interfaces' for 'interface/enabled'"):
        data["interface/enabled"]

    # one multi-level XPath with all keys
    assert data["interface[name='lo']/ietf-ip:ipv6/address[ip='::1']/prefix-length"].as_term().value == 128

    # one item per index, all key values specified
    assert data["interface[name='lo']"]["ietf-ip:ipv6"]["address[ip='::1']"]["prefix-length"].as_term().value == 128

    # one item per index. single-item list keys without their values
    assert data["interface[name='lo']"]["ietf-ip:ipv6"]["address"]["prefix-length"].as_term().value == 128

    assert isinstance(data["interface[name='lo']"]["ietf-ip:ipv6"]["address"]["prefix-length"].as_term().value, int)
